import functools

import numpy as np

from yt.units import dimensions
from yt.units.unit_object import Unit  # type: ignore
from yt.units.unit_registry import UnitRegistry  # type: ignore
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.physical_constants import (
    gravitational_constant_cgs as G,
    speed_of_light_cgs,
)


class Cosmology:
    r"""
    Create a cosmology calculator to compute cosmological distances and times.

    For an explanation of the various cosmological measures, see, for example
    Hogg (1999, https://arxiv.org/abs/astro-ph/9905116).

    WARNING: Cosmological distance calculations return values that are either
    in the comoving or proper frame, depending on the specific quantity.  For
    simplicity, the proper and comoving frames are set equal to each other
    within the cosmology calculator.  This means that for some distance value,
    x, x.to("Mpc") and x.to("Mpccm") will be the same.  The user should take
    care to understand which reference frame is correct for the given calculation.

    Parameters
    ----------
    hubble_constant : float
        The Hubble parameter at redshift zero in units of 100 km/s/Mpc.
        Default: 0.71.
    omega_matter : the fraction of the energy density of the Universe in
        matter at redshift zero.
        Default: 0.27.
    omega_lambda : the fraction of the energy density of the Universe in
        a cosmological constant.
        Default: 0.73.
    omega_radiation : the fraction of the energy density of the Universe in
        relativistic matter at redshift zero.
    omega_curvature : the fraction of the energy density of the Universe in
        curvature.
        Default: 0.0.
    unit_system : :class:`yt.units.unit_systems.UnitSystem`, optional
        The units system to use when making calculations. If not specified,
        cgs units are assumed.
    use_dark_factor: Bool, optional
        The flag to either use the cosmological constant (False, default)
        or to use the parameterization of w(a) as given in Linder 2002. This,
        along with w_0 and w_a, only matters in the function expansion_factor.
    w_0 : float, optional
        The Linder 2002 parameterization of w(a) is: w(a) = w_0 + w_a(1 - a).
        w_0 is w(a = 1). Only matters if use_dark_factor = True. Default is None.
        Cosmological constant case corresponds to w_0 = -1.
    w_a : float, optional
        See w_0. w_a is the derivative of w(a) evaluated at a = 1. Cosmological
        constant case corresponds to w_a = 0. Default is None.

    Examples
    --------

    >>> from yt.utilities.cosmology import Cosmology
    >>> co = Cosmology()
    >>> print(co.t_from_z(0.0).in_units("Gyr"))

    """

    def __init__(
        self,
        hubble_constant=0.71,
        omega_matter=0.27,
        omega_lambda=0.73,
        omega_radiation=0.0,
        omega_curvature=0.0,
        unit_registry=None,
        unit_system="cgs",
        use_dark_factor=False,
        w_0=-1.0,
        w_a=0.0,
    ):
        self.omega_matter = float(omega_matter)
        self.omega_radiation = float(omega_radiation)
        self.omega_lambda = float(omega_lambda)
        self.omega_curvature = float(omega_curvature)
        hubble_constant = float(hubble_constant)
        if unit_registry is None:
            unit_registry = UnitRegistry(unit_system=unit_system)
            unit_registry.add("h", hubble_constant, dimensions.dimensionless, r"h")
            for my_unit in ["m", "pc", "AU", "au"]:
                new_unit = f"{my_unit}cm"
                my_u = Unit(my_unit, registry=unit_registry)
                # technically not true, but distances here are actually comoving
                unit_registry.add(
                    new_unit,
                    my_u.base_value,
                    dimensions.length,
                    "\\rm{%s}/(1+z)" % my_unit,
                    prefixable=True,
                )
        self.unit_registry = unit_registry
        self.hubble_constant = self.quan(hubble_constant, "100*km/s/Mpc")
        self.unit_system = unit_system

        # For non-standard dark energy. If false, use default cosmological constant
        # This only affects the expansion_factor function.
        self.use_dark_factor = use_dark_factor
        self.w_0 = w_0
        self.w_a = w_a

    def hubble_distance(self):
        r"""
        The distance corresponding to c / h, where c is the speed of light
        and h is the Hubble parameter in units of 1 / time.
        """
        return self.quan(speed_of_light_cgs / self.hubble_constant).in_base(
            self.unit_system
        )

    def comoving_radial_distance(self, z_i, z_f):
        r"""
        The comoving distance along the line of sight to on object at redshift,
        z_f, viewed at a redshift, z_i.

        Parameters
        ----------
        z_i : float
            The redshift of the observer.
        z_f : float
            The redshift of the observed object.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.comoving_radial_distance(0.0, 1.0).in_units("Mpccm"))

        """
        return (
            self.hubble_distance() * trapzint(self.inverse_expansion_factor, z_i, z_f)
        ).in_base(self.unit_system)

    def comoving_transverse_distance(self, z_i, z_f):
        r"""
        When multiplied by some angle, the distance between two objects
        observed at redshift, z_f, with an angular separation given by that
        angle, viewed by an observer at redshift, z_i (Hogg 1999).

        Parameters
        ----------
        z_i : float
            The redshift of the observer.
        z_f : float
            The redshift of the observed object.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.comoving_transverse_distance(0.0, 1.0).in_units("Mpccm"))

        """
        if self.omega_curvature > 0:
            return (
                self.hubble_distance()
                / np.sqrt(self.omega_curvature)
                * np.sinh(
                    np.sqrt(self.omega_curvature)
                    * self.comoving_radial_distance(z_i, z_f)
                    / self.hubble_distance()
                )
            ).in_base(self.unit_system)
        elif self.omega_curvature < 0:
            return (
                self.hubble_distance()
                / np.sqrt(np.fabs(self.omega_curvature))
                * np.sin(
                    np.sqrt(np.fabs(self.omega_curvature))
                    * self.comoving_radial_distance(z_i, z_f)
                    / self.hubble_distance()
                )
            ).in_base(self.unit_system)
        else:
            return self.comoving_radial_distance(z_i, z_f)

    def comoving_volume(self, z_i, z_f):
        r"""
        "The comoving volume is the volume measure in which number densities
        of non-evolving objects locked into Hubble flow are constant with
        redshift." -- Hogg (1999)

        Parameters
        ----------
        z_i : float
            The lower redshift of the interval.
        z_f : float
            The higher redshift of the interval.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.comoving_volume(0.0, 1.0).in_units("Gpccm**3"))

        """
        if self.omega_curvature > 1e-10:
            return (
                2
                * np.pi
                * np.power(self.hubble_distance(), 3)
                / self.omega_curvature
                * (
                    self.comoving_transverse_distance(z_i, z_f)
                    / self.hubble_distance()
                    * np.sqrt(
                        1
                        + self.omega_curvature
                        * np.sqrt(
                            self.comoving_transverse_distance(z_i, z_f)
                            / self.hubble_distance()
                        )
                    )
                    - np.sinh(
                        np.fabs(self.omega_curvature)
                        * self.comoving_transverse_distance(z_i, z_f)
                        / self.hubble_distance()
                    )
                    / np.sqrt(self.omega_curvature)
                )
            ).in_base(self.unit_system)
        elif self.omega_curvature < -1e-10:
            return (
                2
                * np.pi
                * np.power(self.hubble_distance(), 3)
                / np.fabs(self.omega_curvature)
                * (
                    self.comoving_transverse_distance(z_i, z_f)
                    / self.hubble_distance()
                    * np.sqrt(
                        1
                        + self.omega_curvature
                        * np.sqrt(
                            self.comoving_transverse_distance(z_i, z_f)
                            / self.hubble_distance()
                        )
                    )
                    - np.arcsin(
                        np.fabs(self.omega_curvature)
                        * self.comoving_transverse_distance(z_i, z_f)
                        / self.hubble_distance()
                    )
                    / np.sqrt(np.fabs(self.omega_curvature))
                )
            ).in_base(self.unit_system)
        else:
            return (
                4 * np.pi * np.power(self.comoving_transverse_distance(z_i, z_f), 3) / 3
            ).in_base(self.unit_system)

    def angular_diameter_distance(self, z_i, z_f):
        r"""
        Following Hogg (1999), the angular diameter distance is 'the ratio of
        an object's physical transverse size to its angular size in radians.'

        Parameters
        ----------
        z_i : float
            The redshift of the observer.
        z_f : float
            The redshift of the observed object.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.angular_diameter_distance(0.0, 1.0).in_units("Mpc"))

        """

        return (
            self.comoving_transverse_distance(0, z_f) / (1 + z_f)
            - self.comoving_transverse_distance(0, z_i) / (1 + z_i)
        ).in_base(self.unit_system)

    def angular_scale(self, z_i, z_f):
        r"""
        The proper transverse distance between two points at redshift z_f
        observed at redshift z_i per unit of angular separation.

        Parameters
        ----------
        z_i : float
            The redshift of the observer.
        z_f : float
            The redshift of the observed object.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.angular_scale(0.0, 1.0).in_units("kpc / arcsec"))

        """

        scale = self.angular_diameter_distance(z_i, z_f) / self.quan(1, "radian")
        return scale.in_base(self.unit_system)

    def luminosity_distance(self, z_i, z_f):
        r"""
        The distance that would be inferred from the inverse-square law of
        light and the measured flux and luminosity of the observed object.

        Parameters
        ----------
        z_i : float
            The redshift of the observer.
        z_f : float
            The redshift of the observed object.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.luminosity_distance(0.0, 1.0).in_units("Mpc"))

        """

        return (
            self.comoving_transverse_distance(0, z_f) * (1 + z_f)
            - self.comoving_transverse_distance(0, z_i) * (1 + z_i)
        ).in_base(self.unit_system)

    def lookback_time(self, z_i, z_f):
        r"""
        The difference in the age of the Universe between the redshift interval
        z_i to z_f.

        Parameters
        ----------
        z_i : float
            The lower redshift of the interval.
        z_f : float
            The higher redshift of the interval.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.lookback_time(0.0, 1.0).in_units("Gyr"))

        """
        return (trapzint(self.age_integrand, z_i, z_f) / self.hubble_constant).in_base(
            self.unit_system
        )

    def critical_density(self, z):
        r"""
        The density required for closure of the Universe at a given
        redshift in the proper frame.

        Parameters
        ----------
        z : float
            Redshift.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.critical_density(0.0).in_units("g/cm**3"))
        >>> print(co.critical_density(0).in_units("Msun/Mpc**3"))

        """
        return (3.0 * self.hubble_parameter(z) ** 2 / 8.0 / np.pi / G).in_base(
            self.unit_system
        )

    def hubble_parameter(self, z):
        r"""
        The value of the Hubble parameter at a given redshift.

        Parameters
        ----------
        z: float
            Redshift.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.hubble_parameter(1.0).in_units("km/s/Mpc"))

        """
        return self.hubble_constant.in_base(self.unit_system) * self.expansion_factor(z)

    def age_integrand(self, z):
        return 1.0 / (z + 1) / self.expansion_factor(z)

    def expansion_factor(self, z):
        r"""
        The ratio between the Hubble parameter at a given redshift and
        redshift zero.

        This is also the primary function integrated to calculate the
        cosmological distances.

        """

        # Use non-standard dark energy
        if self.use_dark_factor:
            dark_factor = self.get_dark_factor(z)

        # Use default cosmological constant
        else:
            dark_factor = 1.0

        zp1 = 1 + z
        return np.sqrt(
            self.omega_matter * zp1**3
            + self.omega_curvature * zp1**2
            + self.omega_radiation * zp1**4
            + self.omega_lambda * dark_factor
        )

    def inverse_expansion_factor(self, z):
        return 1.0 / self.expansion_factor(z)

    def path_length_function(self, z):
        return ((1 + z) ** 2) * self.inverse_expansion_factor(z)

    def path_length(self, z_i, z_f):
        return trapzint(self.path_length_function, z_i, z_f)

    def t_from_a(self, a):
        """
        Compute the age of the Universe for a given scale factor.

        Parameters
        ----------
        a : float
            Scale factor.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.t_from_a(1.0).in_units("Gyr"))

        """

        # Interpolate from a table of log(a) vs. log(t)
        la = np.log10(a)
        la_i = min(-6, np.asarray(la).min() - 3)
        la_f = np.asarray(la).max()
        bins_per_dex = 1000
        n_bins = int((la_f - la_i) * bins_per_dex + 1)
        la_bins = np.linspace(la_i, la_f, n_bins)
        z_bins = 1.0 / np.power(10, la_bins) - 1

        # Integrate in redshift.
        lt = trapezoid_cumulative_integral(self.age_integrand, z_bins)

        # Add a minus sign because we've switched the integration limits.
        table = InterpTable(la_bins[1:], np.log10(-lt))
        t = np.power(10, table(la))

        return (t / self.hubble_constant).in_base(self.unit_system)

    def t_from_z(self, z):
        """
        Compute the age of the Universe for a given redshift.

        Parameters
        ----------
        z : float
            Redshift.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.t_from_z(0.0).in_units("Gyr"))

        """

        return self.t_from_a(1.0 / (1.0 + z))

    def a_from_t(self, t):
        """
        Compute the scale factor for a given age of the Universe.

        Parameters
        ----------
        t : YTQuantity or float
            Time since the Big Bang.  If a float is given, units are
            assumed to be seconds.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.a_from_t(4.0e17))

        """

        if not isinstance(t, YTArray):
            t = self.arr(t, "s")
        lt = np.log10((t * self.hubble_constant).to(""))

        # Interpolate from a table of log(a) vs. log(t)
        # Make initial guess for bounds and widen if necessary.
        la_i = -6
        la_f = 6
        bins_per_dex = 1000
        iter = 0

        while True:
            good = True
            n_bins = int((la_f - la_i) * bins_per_dex + 1)
            la_bins = np.linspace(la_i, la_f, n_bins)
            z_bins = 1.0 / np.power(10, la_bins) - 1

            # Integrate in redshift.
            lt_bins = trapezoid_cumulative_integral(self.age_integrand, z_bins)

            # Add a minus sign because we've switched the integration limits.
            table = InterpTable(np.log10(-lt_bins), la_bins[1:])
            la = table(lt)
            # We want to have the la_bins lower bound be decently
            # below the minimum calculated la values.

            laa = np.asarray(la)
            if laa.min() < la_i + 2:
                la_i -= 3
                good = False
            if laa.max() > la_f:
                la_f = laa.max() + 1
                good = False
            if good:
                break
            iter += 1
            if iter > 10:
                raise RuntimeError("a_from_t calculation did not converge!")

        a = np.power(10, table(lt))
        return a

    def z_from_t(self, t):
        """
        Compute the redshift for a given age of the Universe.

        Parameters
        ----------
        t : YTQuantity or float
            Time since the Big Bang.  If a float is given, units are
            assumed to be seconds.

        Examples
        --------

        >>> from yt.utilities.cosmology import Cosmology
        >>> co = Cosmology()
        >>> print(co.z_from_t(4.0e17))

        """

        a = self.a_from_t(t)
        return 1.0 / a - 1.0

    def get_dark_factor(self, z):
        """
        This function computes the additional term that enters the expansion factor
        when using non-standard dark energy. See Dolag et al 2004 eq. 7 for ref (but
        note that there's a typo in his eq. There should be no negative sign).

        At the moment, this only works using the parameterization given in Linder 2002
        eq. 7: w(a) = w0 + wa(1 - a) = w0 + wa * z / (1+z). This gives rise to an
        analytic expression.
        It is also only functional for Gadget simulations, at the moment.

        Parameters
        ----------
        z:  float
            Redshift
        """

        # Get value of scale factor a corresponding to redshift z
        scale_factor = 1.0 / (1.0 + z)

        # Evaluate exponential using Linder02 parameterization
        dark_factor = np.power(
            scale_factor, -3.0 * (1.0 + self.w_0 + self.w_a)
        ) * np.exp(-3.0 * self.w_a * (1.0 - scale_factor))

        return dark_factor

    _arr = None

    @property
    def arr(self):
        if self._arr is not None:
            return self._arr
        self._arr = functools.partial(YTArray, registry=self.unit_registry)
        return self._arr

    _quan = None

    @property
    def quan(self):
        if self._quan is not None:
            return self._quan
        self._quan = functools.partial(YTQuantity, registry=self.unit_registry)
        return self._quan


def trapzint(f, a, b, bins=10000):
    zbins = np.logspace(np.log10(a + 1), np.log10(b + 1), bins) - 1
    return np.trapz(f(zbins[:-1]), x=zbins[:-1], dx=np.diff(zbins))


def trapezoid_cumulative_integral(f, x):
    """
    Perform cumulative integration using the trapezoid rule.
    """

    fy = f(x)
    return (0.5 * (fy[:-1] + fy[1:]) * np.diff(x)).cumsum()


class InterpTable:
    """
    Generate a function to linearly interpolate from provided arrays.
    """

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __call__(self, val):
        i = np.clip(np.digitize(val, self.x) - 1, 0, self.x.size - 2)
        slope = (self.y[i + 1] - self.y[i]) / (self.x[i + 1] - self.x[i])
        return slope * (val - self.x[i]) + self.y[i]



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.fields.species_fields import setup_species_fields


class SPHFieldInfo(FieldInfoContainer):
    known_particle_fields: KnownFieldsT = (
        ("Mass", ("code_mass", ["particle_mass"], None)),
        ("Masses", ("code_mass", ["particle_mass"], None)),
        ("Coordinates", ("code_length", ["particle_position"], None)),
        ("Velocity", ("code_velocity", ["particle_velocity"], None)),
        ("Velocities", ("code_velocity", ["particle_velocity"], None)),
        ("ParticleIDs", ("", ["particle_index"], None)),
        ("InternalEnergy", ("code_specific_energy", ["specific_thermal_energy"], None)),
        ("SmoothingLength", ("code_length", ["smoothing_length"], None)),
        ("Density", ("code_mass / code_length**3", ["density"], None)),
        ("MaximumTemperature", ("K", [], None)),
        ("Temperature", ("K", ["temperature"], None)),
        ("Epsilon", ("code_length", [], None)),
        ("Metals", ("code_metallicity", ["metallicity"], None)),
        ("Metallicity", ("code_metallicity", ["metallicity"], None)),
        ("Phi", ("code_length", [], None)),
        ("Potential", ("code_velocity**2", ["gravitational_potential"], None)),
        ("StarFormationRate", ("Msun / yr", ["star_formation_rate"], None)),
        ("FormationTime", ("code_time", ["creation_time"], None)),
        ("Metallicity_00", ("", ["metallicity"], None)),
        ("InitialMass", ("code_mass", [], None)),
        ("TrueMass", ("code_mass", [], None)),
        ("ElevenMetalMasses", ("code_mass", [], None)),
        ("ColdFraction", ("", ["cold_fraction"], None)),
        ("HotTemperature", ("code_temperature", ["hot_temperature"], None)),
    )

    def setup_particle_fields(self, ptype, *args, **kwargs):
        super().setup_particle_fields(ptype, *args, **kwargs)
        setup_species_fields(self, ptype)

    def setup_fluid_index_fields(self):
        pass



from collections import defaultdict
from typing import Optional

import numpy as np
from matplotlib.colors import LogNorm, Normalize, SymLogNorm

from yt.funcs import is_sequence, mylog
from yt.units.unit_object import Unit  # type: ignore
from yt.units.yt_array import YTArray
from yt.visualization.plot_container import (
    BaseLinePlot,
    PlotDictionary,
    invalidate_plot,
)


class LineBuffer:
    r"""
    LineBuffer(ds, start_point, end_point, npoints, label = None)

    This takes a data source and implements a protocol for generating a
    'pixelized', fixed-resolution line buffer. In other words, LineBuffer
    takes a starting point, ending point, and number of sampling points and
    can subsequently generate YTArrays of field values along the sample points.

    Parameters
    ----------
    ds : :class:`yt.data_objects.static_output.Dataset`
        This is the dataset object holding the data that can be sampled by the
        LineBuffer
    start_point : n-element list, tuple, ndarray, or YTArray
        Contains the coordinates of the first point for constructing the LineBuffer.
        Must contain n elements where n is the dimensionality of the dataset.
    end_point : n-element list, tuple, ndarray, or YTArray
        Contains the coordinates of the first point for constructing the LineBuffer.
        Must contain n elements where n is the dimensionality of the dataset.
    npoints : int
        How many points to sample between start_point and end_point

    Examples
    --------
    >>> lb = yt.LineBuffer(ds, (0.25, 0, 0), (0.25, 1, 0), 100)
    >>> lb[("all", "u")].max()
    0.11562424257143075 dimensionless

    """

    def __init__(self, ds, start_point, end_point, npoints, label=None):
        self.ds = ds
        self.start_point = _validate_point(start_point, ds, start=True)
        self.end_point = _validate_point(end_point, ds)
        self.npoints = npoints
        self.label = label
        self.data = {}

    def keys(self):
        return self.data.keys()

    def __setitem__(self, item, val):
        self.data[item] = val

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]
        mylog.info("Making a line buffer with %d points of %s", self.npoints, item)
        self.points, self.data[item] = self.ds.coordinates.pixelize_line(
            item, self.start_point, self.end_point, self.npoints
        )

        return self.data[item]

    def __delitem__(self, item):
        del self.data[item]


class LinePlotDictionary(PlotDictionary):
    def __init__(self, data_source):
        super().__init__(data_source)
        self.known_dimensions = {}

    def _sanitize_dimensions(self, item):
        field = self.data_source._determine_fields(item)[0]
        finfo = self.data_source.ds.field_info[field]
        dimensions = Unit(
            finfo.units, registry=self.data_source.ds.unit_registry
        ).dimensions
        if dimensions not in self.known_dimensions:
            self.known_dimensions[dimensions] = item
        return self.known_dimensions[dimensions]

    def __getitem__(self, item):
        ret_item = self._sanitize_dimensions(item)
        return super().__getitem__(ret_item)

    def __setitem__(self, item, value):
        ret_item = self._sanitize_dimensions(item)
        super().__setitem__(ret_item, value)

    def __contains__(self, item):
        ret_item = self._sanitize_dimensions(item)
        return super().__contains__(ret_item)


class LinePlot(BaseLinePlot):
    r"""
    A class for constructing line plots

    Parameters
    ----------

    ds : :class:`yt.data_objects.static_output.Dataset`
        This is the dataset object corresponding to the
        simulation output to be plotted.
    fields : string / tuple, or list of strings / tuples
        The name(s) of the field(s) to be plotted.
    start_point : n-element list, tuple, ndarray, or YTArray
        Contains the coordinates of the first point for constructing the line.
        Must contain n elements where n is the dimensionality of the dataset.
    end_point : n-element list, tuple, ndarray, or YTArray
        Contains the coordinates of the first point for constructing the line.
        Must contain n elements where n is the dimensionality of the dataset.
    npoints : int
        How many points to sample between start_point and end_point for
        constructing the line plot
    figure_size : int or two-element iterable of ints
        Size in inches of the image.
        Default: 5 (5x5)
    fontsize : int
        Font size for all text in the plot.
        Default: 14
    field_labels : dictionary
        Keys should be the field names. Values should be latex-formattable
        strings used in the LinePlot legend
        Default: None


    Example
    -------

    >>> import yt

    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

    >>> plot = yt.LinePlot(ds, "density", [0, 0, 0], [1, 1, 1], 512)
    >>> plot.add_legend("density")
    >>> plot.set_x_unit("cm")
    >>> plot.set_unit("density", "kg/cm**3")
    >>> plot.save()

    """
    _plot_dict_type = LinePlotDictionary
    _plot_type = "line_plot"

    _default_figure_size = (5.0, 5.0)
    _default_font_size = 14.0

    def __init__(
        self,
        ds,
        fields,
        start_point,
        end_point,
        npoints,
        figure_size=None,
        fontsize: Optional[float] = None,
        field_labels=None,
    ):
        """
        Sets up figure and axes
        """
        line = LineBuffer(ds, start_point, end_point, npoints, label=None)
        self.lines = [line]
        self._initialize_instance(self, ds, fields, figure_size, fontsize, field_labels)
        self._setup_plots()

    @classmethod
    def _initialize_instance(
        cls, obj, ds, fields, figure_size, fontsize, field_labels=None
    ):
        obj._x_unit = None
        obj._titles = {}

        data_source = ds.all_data()

        obj.fields = data_source._determine_fields(fields)
        obj.include_legend = defaultdict(bool)
        super(LinePlot, obj).__init__(
            data_source, figure_size=figure_size, fontsize=fontsize
        )
        if field_labels is None:
            obj.field_labels = {}
        else:
            obj.field_labels = field_labels
        for f in obj.fields:
            if f not in obj.field_labels:
                obj.field_labels[f] = f[1]

    def _get_axrect(self):
        fontscale = self._font_properties._size / self.__class__._default_font_size
        top_buff_size = 0.35 * fontscale

        x_axis_size = 1.35 * fontscale
        y_axis_size = 0.7 * fontscale
        right_buff_size = 0.2 * fontscale

        if is_sequence(self.figure_size):
            figure_size = self.figure_size
        else:
            figure_size = (self.figure_size, self.figure_size)

        xbins = np.array([x_axis_size, figure_size[0], right_buff_size])
        ybins = np.array([y_axis_size, figure_size[1], top_buff_size])

        x_frac_widths = xbins / xbins.sum()
        y_frac_widths = ybins / ybins.sum()

        return (
            x_frac_widths[0],
            y_frac_widths[0],
            x_frac_widths[1],
            y_frac_widths[1],
        )

    @classmethod
    def from_lines(
        cls, ds, fields, lines, figure_size=None, font_size=None, field_labels=None
    ):
        """
        A class method for constructing a line plot from multiple sampling lines

        Parameters
        ----------

        ds : :class:`yt.data_objects.static_output.Dataset`
            This is the dataset object corresponding to the
            simulation output to be plotted.
        fields : field name or list of field names
            The name(s) of the field(s) to be plotted.
        lines : list of :class:`yt.visualization.line_plot.LineBuffer` instances
            The lines from which to sample data
        figure_size : int or two-element iterable of ints
            Size in inches of the image.
            Default: 5 (5x5)
        font_size : int
            Font size for all text in the plot.
            Default: 14
        field_labels : dictionary
            Keys should be the field names. Values should be latex-formattable
            strings used in the LinePlot legend
            Default: None

        Example
        --------
        >>> ds = yt.load(
        ...     "SecondOrderTris/RZ_p_no_parts_do_nothing_bcs_cone_out.e", step=-1
        ... )
        >>> fields = [field for field in ds.field_list if field[0] == "all"]
        >>> lines = [
        ...     yt.LineBuffer(ds, [0.25, 0, 0], [0.25, 1, 0], 100, label="x = 0.25"),
        ...     yt.LineBuffer(ds, [0.5, 0, 0], [0.5, 1, 0], 100, label="x = 0.5"),
        ... ]
        >>> lines.append()

        >>> plot = yt.LinePlot.from_lines(ds, fields, lines)
        >>> plot.save()

        """
        obj = cls.__new__(cls)
        obj.lines = lines
        cls._initialize_instance(obj, ds, fields, figure_size, font_size, field_labels)
        obj._setup_plots()
        return obj

    def _setup_plots(self):
        if self._plot_valid:
            return
        for plot in self.plots.values():
            plot.axes.cla()
        for line in self.lines:
            dimensions_counter = defaultdict(int)
            for field in self.fields:
                finfo = self.ds.field_info[field]
                dimensions = Unit(
                    finfo.units, registry=self.ds.unit_registry
                ).dimensions
                dimensions_counter[dimensions] += 1
            for field in self.fields:
                # get plot instance
                plot = self._get_plot_instance(field)

                # calculate x and y
                x, y = self.ds.coordinates.pixelize_line(
                    field, line.start_point, line.end_point, line.npoints
                )

                # scale x and y to proper units
                if self._x_unit is None:
                    unit_x = x.units
                else:
                    unit_x = self._x_unit

                unit_y = plot.norm_handler.display_units

                x.convert_to_units(unit_x)
                y.convert_to_units(unit_y)

                # determine legend label
                str_seq = []
                str_seq.append(line.label)
                str_seq.append(self.field_labels[field])
                delim = "; "
                legend_label = delim.join(filter(None, str_seq))

                # apply plot to matplotlib axes
                plot.axes.plot(x, y, label=legend_label)

                # apply log transforms if requested
                norm = plot.norm_handler.get_norm(data=y)
                y_norm_type = type(norm)
                if y_norm_type is Normalize:
                    plot.axes.set_yscale("linear")
                elif y_norm_type is LogNorm:
                    plot.axes.set_yscale("log")
                elif y_norm_type is SymLogNorm:
                    plot.axes.set_yscale("symlog")
                else:
                    raise NotImplementedError(
                        f"LinePlot doesn't support y norm with type {type(norm)}"
                    )

                # set font properties
                plot._set_font_properties(self._font_properties, None)

                # set x and y axis labels
                axes_unit_labels = self._get_axes_unit_labels(unit_x, unit_y)

                if self._xlabel is not None:
                    x_label = self._xlabel
                else:
                    x_label = r"$\rm{Path\ Length" + axes_unit_labels[0] + "}$"

                if self._ylabel is not None:
                    y_label = self._ylabel
                else:
                    finfo = self.ds.field_info[field]
                    dimensions = Unit(
                        finfo.units, registry=self.ds.unit_registry
                    ).dimensions
                    if dimensions_counter[dimensions] > 1:
                        y_label = (
                            r"$\rm{Multiple\ Fields}$"
                            + r"$\rm{"
                            + axes_unit_labels[1]
                            + "}$"
                        )
                    else:
                        y_label = (
                            finfo.get_latex_display_name()
                            + r"$\rm{"
                            + axes_unit_labels[1]
                            + "}$"
                        )

                plot.axes.set_xlabel(x_label)
                plot.axes.set_ylabel(y_label)

                # apply title
                if field in self._titles:
                    plot.axes.set_title(self._titles[field])

                # apply legend
                dim_field = self.plots._sanitize_dimensions(field)
                if self.include_legend[dim_field]:
                    plot.axes.legend()

        self._plot_valid = True

    @invalidate_plot
    def annotate_legend(self, field):
        """
        Adds a legend to the `LinePlot` instance. The `_sanitize_dimensions`
        call ensures that a legend label will be added for every field of
        a multi-field plot
        """
        dim_field = self.plots._sanitize_dimensions(field)
        self.include_legend[dim_field] = True

    @invalidate_plot
    def set_x_unit(self, unit_name):
        """Set the unit to use along the x-axis

        Parameters
        ----------
        unit_name: str
          The name of the unit to use for the x-axis unit
        """
        self._x_unit = unit_name

    @invalidate_plot
    def set_unit(self, field, new_unit):
        """Set the unit used to plot the field

        Parameters
        ----------
        field: str or field tuple
           The name of the field to set the units for
        new_unit: string or Unit object
        """
        field = self.data_source._determine_fields(field)[0]
        pnh = self.plots[field].norm_handler
        pnh.display_units = new_unit

    @invalidate_plot
    def annotate_title(self, field, title):
        """Set the unit used to plot the field

        Parameters
        ----------
        field: str or field tuple
           The name of the field to set the units for
        title: str
           The title to use for the plot
        """
        self._titles[self.data_source._determine_fields(field)[0]] = title


def _validate_point(point, ds, start=False):
    if not is_sequence(point):
        raise RuntimeError("Input point must be array-like")
    if not isinstance(point, YTArray):
        point = ds.arr(point, "code_length", dtype=np.float64)
    if len(point.shape) != 1:
        raise RuntimeError("Input point must be a 1D array")
    if point.shape[0] < ds.dimensionality:
        raise RuntimeError("Input point must have an element for each dimension")
    # need to pad to 3D elements to avoid issues later
    if point.shape[0] < 3:
        if start:
            val = 0
        else:
            val = 1
        point = np.append(point.d, [val] * (3 - ds.dimensionality)) * point.uq
    return point



from .derived_field import ValidateParameter
from .field_exceptions import NeedsConfiguration, NeedsParameter
from .field_plugin_registry import register_field_plugin


@register_field_plugin
def setup_cosmology_fields(registry, ftype="gas", slice_info=None):
    unit_system = registry.ds.unit_system
    # slice_info would be the left, the right, and the factor.
    # For example, with the old Enzo-ZEUS fields, this would be:
    # slice(None, -2, None)
    # slice(1, -1, None)
    # 1.0
    # Otherwise, we default to a centered difference.
    if slice_info is None:
        sl_left = slice(None, -2, None)
        sl_right = slice(2, None, None)
        div_fac = 2.0
    else:
        sl_left, sl_right, div_fac = slice_info

    def _matter_density(field, data):
        return data[ftype, "density"] + data[ftype, "dark_matter_density"]

    registry.add_field(
        (ftype, "matter_density"),
        sampling_type="local",
        function=_matter_density,
        units=unit_system["density"],
    )

    def _matter_mass(field, data):
        return data[ftype, "matter_density"] * data["index", "cell_volume"]

    registry.add_field(
        (ftype, "matter_mass"),
        sampling_type="local",
        function=_matter_mass,
        units=unit_system["mass"],
    )

    # rho_total / rho_cr(z).
    def _overdensity(field, data):
        if (
            not hasattr(data.ds, "cosmological_simulation")
            or not data.ds.cosmological_simulation
        ):
            raise NeedsConfiguration("cosmological_simulation", 1)
        co = data.ds.cosmology
        return data[ftype, "matter_density"] / co.critical_density(
            data.ds.current_redshift
        )

    registry.add_field(
        (ftype, "overdensity"), sampling_type="local", function=_overdensity, units=""
    )

    # rho_baryon / <rho_baryon>
    def _baryon_overdensity(field, data):
        if (
            not hasattr(data.ds, "cosmological_simulation")
            or not data.ds.cosmological_simulation
        ):
            raise NeedsConfiguration("cosmological_simulation", 1)
        omega_baryon = data.get_field_parameter("omega_baryon")
        if omega_baryon is None:
            raise NeedsParameter("omega_baryon")
        co = data.ds.cosmology
        # critical_density(z) ~ omega_lambda + omega_matter * (1 + z)^3
        # mean matter density(z) ~ omega_matter * (1 + z)^3
        return (
            data[ftype, "density"]
            / omega_baryon
            / co.critical_density(0.0)
            / (1.0 + data.ds.current_redshift) ** 3
        )

    registry.add_field(
        (ftype, "baryon_overdensity"),
        sampling_type="local",
        function=_baryon_overdensity,
        units="",
        validators=[ValidateParameter("omega_baryon")],
    )

    # rho_matter / <rho_matter>
    def _matter_overdensity(field, data):
        if (
            not hasattr(data.ds, "cosmological_simulation")
            or not data.ds.cosmological_simulation
        ):
            raise NeedsConfiguration("cosmological_simulation", 1)
        co = data.ds.cosmology
        # critical_density(z) ~ omega_lambda + omega_matter * (1 + z)^3
        # mean density(z) ~ omega_matter * (1 + z)^3
        return (
            data[ftype, "matter_density"]
            / data.ds.omega_matter
            / co.critical_density(0.0)
            / (1.0 + data.ds.current_redshift) ** 3
        )

    registry.add_field(
        (ftype, "matter_overdensity"),
        sampling_type="local",
        function=_matter_overdensity,
        units="",
    )

    # r / r_vir
    def _virial_radius_fraction(field, data):
        virial_radius = data.get_field_parameter("virial_radius")
        if virial_radius == 0.0:
            ret = 0.0
        else:
            ret = data[("index", "radius")] / virial_radius
        return ret

    registry.add_field(
        ("index", "virial_radius_fraction"),
        sampling_type="local",
        function=_virial_radius_fraction,
        validators=[ValidateParameter("virial_radius")],
        units="",
    )

    # Weak lensing convergence.
    # Eqn 4 of Metzler, White, & Loken (2001, ApJ, 547, 560).
    # This needs to be checked for accuracy.
    def _weak_lensing_convergence(field, data):
        if (
            not hasattr(data.ds, "cosmological_simulation")
            or not data.ds.cosmological_simulation
        ):
            raise NeedsConfiguration("cosmological_simulation", 1)
        co = data.ds.cosmology
        pc = data.ds.units.physical_constants
        observer_redshift = data.get_field_parameter("observer_redshift")
        source_redshift = data.get_field_parameter("source_redshift")

        # observer to lens
        dl = co.angular_diameter_distance(observer_redshift, data.ds.current_redshift)
        # observer to source
        ds = co.angular_diameter_distance(observer_redshift, source_redshift)
        # lens to source
        dls = co.angular_diameter_distance(data.ds.current_redshift, source_redshift)

        # removed the factor of 1 / a to account for the fact that we are projecting
        # with a proper distance.
        return (
            1.5
            * (co.hubble_constant / pc.clight) ** 2
            * (dl * dls / ds)
            * data[ftype, "matter_overdensity"]
        ).in_units("1/cm")

    registry.add_field(
        (ftype, "weak_lensing_convergence"),
        sampling_type="local",
        function=_weak_lensing_convergence,
        units=unit_system["length"] ** -1,
        validators=[
            ValidateParameter("observer_redshift"),
            ValidateParameter("source_redshift"),
        ],
    )



import numpy as np

from yt.funcs import mylog, parse_h5_attr
from yt.geometry.selection_routines import GridSelector
from yt.units.yt_array import uvstack  # type: ignore
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.on_demand_imports import _h5py as h5py


class IOHandlerYTNonspatialhdf5(BaseIOHandler):
    _dataset_type = "ytnonspatialhdf5"
    _base = slice(None)
    _field_dtype = "float64"

    def _read_fluid_selection(self, g, selector, fields):
        rv = {}
        if isinstance(selector, GridSelector):
            if g.id in self._cached_fields:
                gf = self._cached_fields[g.id]
                rv.update(gf)
            if len(rv) == len(fields):
                return rv
            f = h5py.File(g.filename, mode="r")
            for field in fields:
                if field in rv:
                    self._hits += 1
                    continue
                self._misses += 1
                ftype, fname = field
                rv[(ftype, fname)] = f[ftype][fname][()]
            if self._cache_on:
                for gid in rv:
                    self._cached_fields.setdefault(gid, {})
                    self._cached_fields[gid].update(rv[gid])
            f.close()
            return rv
        else:
            raise RuntimeError(
                "Geometric selection not supported for non-spatial datasets."
            )


class IOHandlerYTGridHDF5(BaseIOHandler):
    _dataset_type = "ytgridhdf5"
    _base = slice(None)
    _field_dtype = "float64"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        rv = {}
        # Now we have to do something unpleasant
        chunks = list(chunks)
        if isinstance(selector, GridSelector):
            if not (len(chunks) == len(chunks[0].objs) == 1):
                raise RuntimeError
            g = chunks[0].objs[0]
            if g.id in self._cached_fields:
                gf = self._cached_fields[g.id]
                rv.update(gf)
            if len(rv) == len(fields):
                return rv
            f = h5py.File(g.filename, mode="r")
            gds = f[self.ds.default_fluid_type]
            for field in fields:
                if field in rv:
                    self._hits += 1
                    continue
                self._misses += 1
                ftype, fname = field
                rv[(ftype, fname)] = gds[fname][()]
            if self._cache_on:
                for gid in rv:
                    self._cached_fields.setdefault(gid, {})
                    self._cached_fields[gid].update(rv[gid])
            f.close()
            return rv
        if size is None:
            size = sum(g.count(selector) for chunk in chunks for g in chunk.objs)
        for field in fields:
            ftype, fname = field
            fsize = size
            rv[field] = np.empty(fsize, dtype="float64")
        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s grids",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )
        ind = 0
        for chunk in chunks:
            f = None
            for g in chunk.objs:
                if g.filename is None:
                    continue
                if f is None:
                    f = h5py.File(g.filename, mode="r")
                gf = self._cached_fields.get(g.id, {})
                nd = 0
                for field in fields:
                    if field in gf:
                        nd = g.select(selector, gf[field], rv[field], ind)
                        self._hits += 1
                        continue
                    self._misses += 1
                    ftype, fname = field
                    # add extra dimensions to make data 3D
                    data = f[ftype][fname][()].astype(self._field_dtype)
                    for dim in range(len(data.shape), 3):
                        data = np.expand_dims(data, dim)
                    if self._cache_on:
                        self._cached_fields.setdefault(g.id, {})
                        self._cached_fields[g.id][field] = data
                    nd = g.select(selector, data, rv[field], ind)  # caches
                ind += nd
            if f:
                f.close()
        return rv

    def _read_particle_coords(self, chunks, ptf):
        pn = "particle_position_%s"
        chunks = list(chunks)
        for chunk in chunks:
            f = None
            for g in chunk.objs:
                if g.filename is None:
                    continue
                if f is None:
                    f = h5py.File(g.filename, mode="r")
                if g.NumberOfParticles == 0:
                    continue
                for ptype, field_list in sorted(ptf.items()):
                    units = parse_h5_attr(f[ptype][pn % "x"], "units")
                    x, y, z = (
                        self.ds.arr(f[ptype][pn % ax][()].astype("float64"), units)
                        for ax in "xyz"
                    )
                    for field in field_list:
                        if np.asarray(f[ptype][field]).ndim > 1:
                            self._array_fields[field] = f[ptype][field].shape[1:]
                    yield ptype, (x, y, z), 0.0
            if f:
                f.close()

    def _read_particle_fields(self, chunks, ptf, selector):
        pn = "particle_position_%s"
        chunks = list(chunks)
        for chunk in chunks:  # These should be organized by grid filename
            f = None
            for g in chunk.objs:
                if g.filename is None:
                    continue
                if f is None:
                    f = h5py.File(g.filename, mode="r")
                if g.NumberOfParticles == 0:
                    continue
                for ptype, field_list in sorted(ptf.items()):
                    units = parse_h5_attr(f[ptype][pn % "x"], "units")
                    x, y, z = (
                        self.ds.arr(f[ptype][pn % ax][()].astype("float64"), units)
                        for ax in "xyz"
                    )
                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None:
                        continue
                    for field in field_list:
                        data = np.asarray(f[ptype][field][()], "=f8")
                        yield (ptype, field), data[mask]
            if f:
                f.close()


class IOHandlerYTDataContainerHDF5(BaseIOHandler):
    _dataset_type = "ytdatacontainer_hdf5"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _yield_coordinates(self, data_file):
        with h5py.File(data_file.filename, mode="r") as f:
            for ptype in f.keys():
                if "x" not in f[ptype].keys():
                    continue
                units = _get_position_array_units(ptype, f, "x")
                x, y, z = (
                    self.ds.arr(_get_position_array(ptype, f, ax), units)
                    for ax in "xyz"
                )
                pos = uvstack([x, y, z]).T
                pos.convert_to_units("code_length")
                yield ptype, pos

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype in sorted(ptf):
                    pcount = data_file.total_particles[ptype]
                    if pcount == 0:
                        continue
                    units = _get_position_array_units(ptype, f, "x")
                    x, y, z = (
                        self.ds.arr(_get_position_array(ptype, f, ax), units)
                        for ax in "xyz"
                    )
                    yield ptype, (x, y, z), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype, field_list in sorted(ptf.items()):
                    units = _get_position_array_units(ptype, f, "x")
                    x, y, z = (
                        self.ds.arr(_get_position_array(ptype, f, ax), units)
                        for ax in "xyz"
                    )
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    if mask is None:
                        continue
                    for field in field_list:
                        data = f[ptype][field][mask].astype("float64")
                        yield (ptype, field), data

    def _count_particles(self, data_file):
        si, ei = data_file.start, data_file.end
        if None not in (si, ei):
            pcount = {}
            for ptype, npart in self.ds.num_particles.items():
                pcount[ptype] = np.clip(npart - si, 0, ei - si)
        else:
            pcount = self.ds.num_particles
        return pcount

    def _identify_fields(self, data_file):
        fields = []
        units = {}
        with h5py.File(data_file.filename, mode="r") as f:
            for ptype in f:
                fields.extend([(ptype, str(field)) for field in f[ptype]])
                units.update(
                    {
                        (ptype, str(field)): parse_h5_attr(f[ptype][field], "units")
                        for field in f[ptype]
                    }
                )
        return fields, units


class IOHandlerYTSpatialPlotHDF5(IOHandlerYTDataContainerHDF5):
    _dataset_type = "ytspatialplot_hdf5"

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype in sorted(ptf):
                    pcount = data_file.total_particles[ptype]
                    if pcount == 0:
                        continue
                    x = _get_position_array(ptype, f, "px")
                    y = _get_position_array(ptype, f, "py")
                    z = (
                        np.zeros(x.size, dtype="float64")
                        + self.ds.domain_left_edge[2].to("code_length").d
                    )
                    yield ptype, (x, y, z), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        chunks = list(chunks)
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            all_count = self._count_particles(data_file)
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype, field_list in sorted(ptf.items()):
                    x = _get_position_array(ptype, f, "px")
                    y = _get_position_array(ptype, f, "py")
                    z = (
                        np.zeros(all_count[ptype], dtype="float64")
                        + self.ds.domain_left_edge[2].to("code_length").d
                    )
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    if mask is None:
                        continue
                    for field in field_list:
                        data = f[ptype][field][mask].astype("float64")
                        yield (ptype, field), data


def _get_position_array(ptype, f, ax):
    if ptype == "grid":
        pos_name = ""
    else:
        pos_name = "particle_position_"
    return f[ptype][pos_name + ax][()].astype("float64")


def _get_position_array_units(ptype, f, ax):
    if ptype == "grid":
        pos_name = ""
    else:
        pos_name = "particle_position_"
    return parse_h5_attr(f[ptype][pos_name + ax], "units")
